/* 
 * Copyright (C) 2016 Du-Lab Team <dulab.binf@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package dulab.adap.common.distances;

import dulab.adap.common.types.SparseVector;

import java.io.Serializable;

import java.util.Iterator;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.NoSuchElementException;
import java.util.TreeMap;

/**
 *
 * @author aleksandrsmirnov
 */
public class PurityScore extends Distance implements Serializable
{
    private static final double SCALE = 2 / java.lang.Math.PI;
    private static final double TOLERANCE = 0.5; // 1e-3;
    
    @Override
    public final double call(SparseVector v1, SparseVector v2) {
        return call(v1, v2, TOLERANCE);
    }
    
    @Override
    public final double call(SparseVector v1, SparseVector v2, double tolerance)
    {
        if (v1.size() == 0 || v2.size() == 0) return 1.0;
        
        v1 = scale(v1);
        v2 = scale(v2);
        
        double sum = 0.0;
        
        Iterator <Entry <Double, Double>> it1 = v1.data().entrySet().iterator();
        Iterator <Entry <Double, Double>> it2 = v2.data().entrySet().iterator();
        
        Entry <Double, Double> e1 = it1.next();
        Entry <Double, Double> e2 = it2.next();
        
        for (;;)
        {
            double delta = e1.getKey() - e2.getKey();

            try {
                if (delta < -tolerance) // x1 < x2 - tolerance
                    e1 = it1.next();
                else if (delta > tolerance) // x2 < x1 - tolerance
                    e2 = it2.next();
                else { // |x1 - x2| <= tolerance
                    sum += e1.getValue() * e2.getValue();
                    e1 = it1.next();
                    e2 = it2.next();
                }
            } catch (NoSuchElementException e) {break;}
        }
        
        sum *= purityFactor(v1, v2, tolerance); // Normalize
        
        if (sum > 1.0) sum = 1.0; // If sum > 1 then Math.acos(sum) gives NaN
        
        return SCALE * Math.acos(sum * sum);
    }
    
    private SparseVector scale(SparseVector v) {
        NavigableMap <Double, Double> data = new TreeMap <> (v.data());
        
        // Scale vector to have the maximum intensity equals to 1
        double maxIntensity = 0.0;
        for (double intensity : data.values())
            if (intensity > maxIntensity)
                maxIntensity = intensity;
        
        double scale = 1.0 / maxIntensity;
        double totalIntensity = 0.0;
        
        for (Entry <Double, Double> e : data.entrySet()) {
            double newIntensity = scale * e.getValue();
            e.setValue(newIntensity);
            totalIntensity += newIntensity;
        }
        
        double omega = 1.0 / (totalIntensity - 0.5);
        
        // Scale based on m/z-values and total intensity
        for (Entry <Double, Double> e : data.entrySet()) {
            double intensity = e.getValue();
            double newIntensity = Math.sqrt(intensity * e.getKey() 
                    / (1.0 + omega * intensity));
            e.setValue(newIntensity);
        }
        
        return new SparseVector(data);
    }
    
    double purityFactor(SparseVector v1, SparseVector v2, double tolerance) 
    {   
        Iterator <Entry <Double, Double>> it1 = v1.data().entrySet().iterator();
        Iterator <Entry <Double, Double>> it2 = v2.data().entrySet().iterator();
        
        Entry <Double, Double> e1 = it1.next();
        Entry <Double, Double> e2 = it2.next();
        
        double sum1 = 0.0, sum2 = 0.0;
        
        for (;;)
        {
            double delta = e1.getKey() - e2.getKey();

            try {
                if (delta < -tolerance) // x1 < x2 - tolerance
                    e1 = it1.next();
                else if (delta > tolerance) // x2 < x1 - tolerance
                    e2 = it2.next();
                else { // |x1 - x2| <= tolerance
                    sum1 += e1.getValue() * e1.getValue();
                    sum2 += e2.getValue() * e2.getValue();
                    e1 = it1.next();
                    e2 = it2.next();
                }
            } catch (NoSuchElementException e) {break;}
        }
               
        // return (0.3 / v1.norm() + 0.7 / Math.sqrt(sum1)) / v2.norm();
        
        return 0.3 / v1.norm() / v2.norm() 
                + 0.7 / Math.sqrt(sum1) / Math.sqrt(sum2);
    }
}
